@@ -1,4 +1,4 @@
-# Copyright 2020 The Trieste Contributors
+# Copyright 2021 The Trieste Contributors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,16 +16,18 @@ from __future__ import annotations
 import copy
 from abc import ABC, abstractmethod
 from collections.abc import Callable
-from typing import Any, TypeVar
+from typing import Any, Optional, TypeVar
 
 import gpflow
 import tensorflow as tf
+import tensorflow_probability as tfp
 from gpflow.models import GPR, SGPR, SVGP, VGP, GPModel
+from gpflow.utilities import multiple_assign, read_values
 
 from ..data import Dataset
 from ..type import TensorType
-from ..utils import DEFAULTS
-from .optimizer import Optimizer
+from ..utils import DEFAULTS, jit
+from .optimizer import Optimizer, TFOptimizer
 
 
 class ProbabilisticModel(ABC):
@@ -87,6 +89,24 @@ class ProbabilisticModel(ABC):
             f"Model {self!r} does not support predicting observations, just the latent function"
         )
 
+    def get_observation_noise(self) -> TensorType:
+        """
+        Return the variance of observation noise.
+
+        Note that this is not supported by all models.
+
+        :return: The observation noise.
+        """
+        raise NotImplementedError(f"Model {self!r} does not provide scalar observation noise")
+
+    def get_kernel(self) -> gpflow.kernels.Kernel:
+        """
+        Return the kernel of the model.
+
+        :return: The kernel.
+        """
+        raise NotImplementedError("Model {self!r} does not provide observation noise")
+
 
 class TrainableProbabilisticModel(ProbabilisticModel):
     """A trainable probabilistic model."""
@@ -295,6 +315,27 @@ class GPflowPredictor(ProbabilisticModel, tf.Module, ABC):
     def predict_y(self, query_points: TensorType) -> tuple[TensorType, TensorType]:
         return self.model.predict_y(query_points)
 
+    def get_kernel(self) -> gpflow.kernels.Kernel:
+        """
+        Return the kernel of the model.
+
+        :return: The kernel.
+        """
+        return self.model.kernel
+
+    def get_observation_noise(self):
+        """
+        Return the variance of observation noise for homoscedastic likelihoods.
+        :return: The observation noise.
+        :raise NotImplementedError: If the model does not have a homoscedastic likelihood.
+        """
+        try:
+            noise_variance = self.model.likelihood.variance
+        except AttributeError:
+            raise NotImplementedError("Model {self!r} does not have scalar observation noise")
+
+        return noise_variance
+
     def optimize(self, dataset: Dataset) -> None:
         """
         Optimize the model with the specified `dataset`.
@@ -312,15 +353,26 @@ class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel):
     or :class:`~gpflow.models.SGPR`.
     """
 
-    def __init__(self, model: GPR | SGPR, optimizer: Optimizer | None = None):
+    def __init__(
+        self, model: GPR | SGPR, optimizer: Optimizer | None = None, num_kernel_samples: int = 10
+    ):
         """
         :param model: The GPflow model to wrap.
         :param optimizer: The optimizer with which to train the model. Defaults to
             :class:`~trieste.models.optimizer.Optimizer` with :class:`~gpflow.optimizers.Scipy`.
+        :param num_kernel_samples: Number of randomly sampled kernels (for each kernel parameter) to
+            evaluate before beginning model optimization. Therefore, for a kernel with `p`
+            (vector-valued) parameters, we evaluate `p * num_kernel_samples` kernels.
         """
         super().__init__(optimizer)
         self._model = model
 
+        if num_kernel_samples <= 0:
+            raise ValueError(
+                f"num_kernel_samples must be greater or equal to zero but got {num_kernel_samples}."
+            )
+        self._num_kernel_samples = num_kernel_samples
+
     def __repr__(self) -> str:
         """"""
         return f"GaussianProcessRegression({self._model!r}, {self.optimizer!r})"
@@ -379,6 +431,73 @@ class GaussianProcessRegression(GPflowPredictor, TrainableProbabilisticModel):
 
         return cov
 
+    def optimize(self, dataset: Dataset) -> None:
+        """
+        Optimize the model with the specified `dataset`.
+
+        For :class:`GaussianProcessRegression`, we (optionally) try multiple randomly sampled
+        kernel parameter configurations as well as the configuration specified when initializing
+        the kernel. The best configuration is used as the starting point for model optimization.
+
+        For trainable parameters constrained to lie in a finite interval (through a sigmoid
+        bijector), we begin model optimization from the best of a random sample from these
+        parameters' acceptable domains.
+
+        For trainable parameters without constraints but with priors, we begin model optimization
+        from the best of a random sample from these parameters' priors.
+
+        For trainable parameters with neither priors nor constraints, we begin optimization from
+        their initial values.
+
+        :param dataset: The data with which to optimize the `model`.
+        """
+
+        num_trainable_params_with_priors_or_constraints = tf.reduce_sum(
+            [
+                tf.size(param)
+                for param in self.model.trainable_parameters
+                if param.prior is not None or isinstance(param.bijector, tfp.bijectors.Sigmoid)
+            ]
+        )
+
+        if (
+            min(num_trainable_params_with_priors_or_constraints, self._num_kernel_samples) >= 1
+        ):  # Find a promising kernel initialization
+            self.find_best_model_initialization(
+                self._num_kernel_samples * num_trainable_params_with_priors_or_constraints
+            )
+
+        self.optimizer.optimize(self.model, dataset)
+
+    def find_best_model_initialization(self, num_kernel_samples) -> None:
+        """
+        Test `num_kernel_samples` models with sampled kernel parameters. The model's kernel
+        parameters are then set to the sample achieving maximal likelihood.
+
+        :param num_kernel_samples: Number of randomly sampled kernels to evaluate.
+        """
+
+        @tf.function
+        def evaluate_loss_of_model_parameters() -> tf.Tensor:
+            randomize_hyperparameters(self.model)
+            return self.model.training_loss()
+
+        squeeze_hyperparameters(self.model)
+        current_best_parameters = read_values(self.model)
+        min_loss = self.model.training_loss()
+
+        for _ in tf.range(num_kernel_samples):
+            try:
+                train_loss = evaluate_loss_of_model_parameters()
+            except tf.errors.InvalidArgumentError:  # allow badly specified kernel params
+                train_loss = 1e100
+
+            if train_loss < min_loss:  # only keep best kernel params
+                min_loss = train_loss
+                current_best_parameters = read_values(self.model)
+
+        multiple_assign(self.model, current_best_parameters)
+
 
 class SparseVariational(GPflowPredictor, TrainableProbabilisticModel):
     """
@@ -414,19 +533,65 @@ class SparseVariational(GPflowPredictor, TrainableProbabilisticModel):
 
 
 class VariationalGaussianProcess(GPflowPredictor, TrainableProbabilisticModel):
-    """A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.VGP`."""
+    r"""
+    A :class:`TrainableProbabilisticModel` wrapper for a GPflow :class:`~gpflow.models.VGP`.
 
-    def __init__(self, model: VGP, optimizer: Optimizer | None = None):
+    A Variational Gaussian Process (VGP) approximates the posterior of a GP
+    using the multivariate Gaussian closest to the posterior of the GP by minimizing the
+    KL divergence between approximated and exact posteriors. See :cite:`opper2009variational`
+    for details.
+
+    The VGP provides (approximate) GP modelling under non-Gaussian likelihoods, for example
+    when fitting a classification model over binary data.
+
+    A whitened representation and (optional) natural gradient steps are used to aid
+    model optimization.
+    """
+
+    def __init__(
+        self,
+        model: VGP,
+        optimizer: Optimizer | None = None,
+        use_natgrads: bool = False,
+        natgrad_gamma: Optional[float] = None,
+    ):
         """
         :param model: The GPflow :class:`~gpflow.models.VGP`.
         :param optimizer: The optimizer with which to train the model. Defaults to
             :class:`~trieste.models.optimizer.Optimizer` with :class:`~gpflow.optimizers.Scipy`.
-        :raise ValueError (or InvalidArgumentError): If ``model``'s :attr:`q_sqrt` is not rank 3.
+        :param use_natgrads: If True then alternate model optimization steps with natural
+            gradient updates. Note that natural gradients requires
+            an :class:`~trieste.models.optimizer.Optimizer` optimizer.
+        :natgrad_gamma: Gamma parameter for the natural gradient optimizer.
+        :raise ValueError (or InvalidArgumentError): If ``model``'s :attr:`q_sqrt` is not rank 3
+            or if attempting to combine natural gradients with a :class:`~gpflow.optimizers.Scipy`
+            optimizer.
         """
         tf.debugging.assert_rank(model.q_sqrt, 3)
         super().__init__(optimizer)
         self._model = model
 
+        if use_natgrads:
+            if not isinstance(self._optimizer, TFOptimizer):
+                raise ValueError(
+                    f"""
+                    Natgrads can only be used alongside an optimizer from tf.optimizers however
+                    received f{self._optimizer}
+                    """
+                )
+
+            natgrad_gamma = 0.1 if natgrad_gamma is None else natgrad_gamma
+        else:
+            if natgrad_gamma is not None:
+                raise ValueError(
+                    """
+                    natgrad_gamma is only to be specified when use_natgrads is True.
+                    """
+                )
+
+        self._use_natgrads = use_natgrads
+        self._natgrad_gamma = natgrad_gamma
+
     def __repr__(self) -> str:
         """"""
         return f"VariationalGaussianProcess({self._model!r}, {self.optimizer!r})"
@@ -440,7 +605,7 @@ class VariationalGaussianProcess(GPflowPredictor, TrainableProbabilisticModel):
         Update the model given the specified ``dataset``. Does not train the model.
 
         :param dataset: The data with which to update the model.
-        :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
+        :param jitter: The size of the jitter to use when stabilizing the Cholesky decomposition of
             the covariance matrix.
         """
         model = self.model
@@ -466,6 +631,48 @@ class VariationalGaussianProcess(GPflowPredictor, TrainableProbabilisticModel):
         model.q_mu = gpflow.Parameter(new_q_mu)
         model.q_sqrt = gpflow.Parameter(new_q_sqrt, transform=gpflow.utilities.triangular())
 
+    def optimize(self, dataset: Dataset) -> None:
+        """
+        :class:`VariationalGaussianProcess` has a custom `optimize` method that (optionally) permits
+        alternating between standard optimization steps (for kernel parameters) and natural gradient
+        steps for the variational parameters (`q_mu` and `q_sqrt`). See :cite:`salimbeni2018natural`
+        for details. Using natural gradients can dramatically speed up model fitting, especially for
+        ill-conditioned posteriors.
+
+        If using natural gradients, our optimizer inherits the mini-batch behavior and number
+        of optimization steps as the base optimizer specified when initializing
+        the :class:`VariationalGaussianProcess`.
+        """
+        model = self.model
+
+        if self._use_natgrads:  # optimize variational params with natgrad optimizer
+
+            natgrad_optimizer = gpflow.optimizers.NaturalGradient(gamma=self._natgrad_gamma)
+            base_optimizer = self.optimizer
+
+            gpflow.set_trainable(model.q_mu, False)  # variational params optimized by natgrad
+            gpflow.set_trainable(model.q_sqrt, False)
+            variational_params = [(model.q_mu, model.q_sqrt)]
+            model_params = model.trainable_variables
+
+            loss_fn = base_optimizer.create_loss(model, dataset)
+
+            @jit(apply=self.optimizer.compile)
+            def perfom_optimization_step() -> None:  # alternate with natgrad optimizations
+                natgrad_optimizer.minimize(loss_fn, variational_params)
+                base_optimizer.optimizer.minimize(
+                    loss_fn, model_params, **base_optimizer.minimize_args
+                )
+
+            for _ in range(base_optimizer.max_iter):  # type: ignore
+                perfom_optimization_step()
+
+            gpflow.set_trainable(model.q_mu, True)  # revert varitional params to trainable
+            gpflow.set_trainable(model.q_sqrt, True)
+
+        else:
+            self.optimizer.optimize(model, dataset)
+
 
 supported_models: dict[Any, Callable[[Any, Optimizer], TrainableProbabilisticModel]] = {
     GPR: GaussianProcessRegression,
@@ -492,3 +699,63 @@ def _assert_data_is_compatible(new_data: Dataset, existing_data: Dataset) -> Non
             f" shape {existing_data.observations.shape} of existing observations. Trailing"
             f" dimensions must match."
         )
+
+
+def randomize_hyperparameters(object: gpflow.Module) -> None:
+    """
+    Sets hyperparameters to random samples from their constrained domains or (if not constraints
+    are available) their prior distributions.
+
+    :param object: Any gpflow Module.
+    """
+    for param in object.trainable_parameters:
+        if isinstance(param.bijector, tfp.bijectors.Sigmoid):
+            sample = tf.random.uniform(
+                param.bijector.low.shape,
+                minval=param.bijector.low,
+                maxval=param.bijector.high,
+                dtype=param.bijector.low.dtype,
+            )
+            param.assign(sample)
+        elif param.prior is not None:
+            param.assign(param.prior.sample())
+
+
+def squeeze_hyperparameters(
+    object: gpflow.Module, alpha: float = 1e-2, epsilon: float = 1e-7
+) -> None:
+    """
+    Squeezes the parameters to be strictly inside their range defined by the Sigmoid,
+    or strictly greater than the limit defined by the Shift+Softplus.
+    This avoids having Inf unconstrained values when the parameters are exactly at the boundary.
+
+    :param object: Any gpflow Module.
+    :param alpha: the proportion of the range with which to squeeze for the Sigmoid case
+    :param epsilon: the value with which to offset the shift for the Softplus case.
+    :raise ValueError: If ``alpha`` is not in (0,1) or epsilon <= 0
+    """
+
+    if not (0 < alpha < 1):
+        raise ValueError(f"squeeze factor alpha must be in (0, 1), found {alpha}")
+
+    if not (0 < epsilon):
+        raise ValueError(f"offset factor epsilon must be > 0, found {epsilon}")
+
+    for param in object.trainable_parameters:
+        if isinstance(param.bijector, tfp.bijectors.Sigmoid):
+            delta = (param.bijector.high - param.bijector.low) * alpha
+            squeezed_param = tf.math.minimum(param, param.bijector.high - delta)
+            squeezed_param = tf.math.maximum(squeezed_param, param.bijector.low + delta)
+            param.assign(squeezed_param)
+        elif (
+            isinstance(param.bijector, tfp.bijectors.Chain)
+            and len(param.bijector.bijectors) == 2
+            and isinstance(param.bijector.bijectors[0], tfp.bijectors.Shift)
+            and isinstance(param.bijector.bijectors[1], tfp.bijectors.Softplus)
+        ):
+            if isinstance(param.bijector.bijectors[0], tfp.bijectors.Shift) and isinstance(
+                param.bijector.bijectors[1], tfp.bijectors.Softplus
+            ):
+                low = param.bijector.bijectors[0].shift
+                squeezed_param = tf.math.maximum(param, low + epsilon * tf.ones_like(param))
+                param.assign(squeezed_param)
